'From Squeak3.9 of 7 November 2006 [latest update: #7067] on 25 March 2009 at 10:46:44 pm'!
"Change Set:		FilesImprovements
Date:			19 February 2009
Author:			Chris Muller

A refactoring of DirectoryEntry to no longer employ an Array for its internal storage.  Also adds several fixes and enhancements."
"Define subclass just to ensure this can file-in!!"
DirectoryEntry subclass: #DirectoryEntryDirectory
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Directories'.
DirectoryEntry subclass: #DirectoryEntryFile
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Directories'!

Object subclass: #DirectoryEntry
	instanceVariableNames: 'directory name creationTime modificationTime fileSize attributes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Directories'!

!DirectoryEntry commentStamp: '<historical>' prior: 0!
an entry in a directory; a reference to either a file or a directory.!


!DirectoryEntryDirectory commentStamp: 'cmm 9/13/2007 12:24' prior: 0!
an entry in a directory; a reference to a directory.!


!DirectoryEntryFile commentStamp: 'cmm 9/13/2007 12:24' prior: 0!
an entry in a directory; a reference to a file.!


!DirectoryEntry methodsFor: 'access' stamp: 'cmm 8/22/2007 11:59'!
at: index
	"compatibility interface"
	self halt: 'old-style access to DirectoryEntry'.
	index = 1 ifTrue: [ ^self name ].
	index = 2 ifTrue: [ ^self creationTime ].
	index = 3 ifTrue: [ ^self modificationTime ].
	index = 4 ifTrue:[ ^self isDirectory ].
	index = 5 ifTrue:[ ^self fileSize ].
	self error: 'invalid index specified'.
! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 9/13/2007 16:54'!
attributeNamed: aString ifAbsent: oneArgBlock 
	"Answer the value of attribute named aString.  If no attribute by that name has ever been set, answer oneArgBlock value."
	^ self attributes 
		at: aString
		ifAbsent: oneArgBlock! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 9/13/2007 16:55'!
attributeNamed: aString put: anObject
	"Set the value of attribute named aString.  This is provided for extension by third-party developers."
	^ self attributes 
		at: aString
		put: anObject! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 9/12/2007 17:36'!
containingDirectory
	"Answer the FileDirectory in which I reside."
	^ directory! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 8/10/2007 12:25'!
creationDateAndTime
	"The DateAndTime my entry in the file system was created."
	^DateAndTime fromSeconds: creationTime! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 11/30/2008 23:48'!
fullName
	"The fully-qualified name."
	^ directory fullNameFor: self name! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 9/12/2007 17:25'!
isDirectory
	"whether this entry represents a directory"
	self subclassResponsibility! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 8/10/2007 12:25'!
modificationDateAndTime
	"The DateAndTime my entry in the file system was last modified."
	^ DateAndTime fromSeconds: modificationTime! !

!DirectoryEntry methodsFor: 'access' stamp: 'cmm 8/29/2007 17:44'!
printOn: aStream 
	super printOn: aStream.
	aStream
		space ;
		nextPutAll: self name! !

!DirectoryEntry methodsFor: 'file operations' stamp: 'cmm 9/26/2007 13:36'!
copyTo: fileOrServerDirectory 
	"Copy me to fileOrServerDirectory."
	self subclassResponsibility! !

!DirectoryEntry methodsFor: 'file operations' stamp: 'cmm 9/26/2007 13:03'!
delete
	"Physically remove from the disk."
	self subclassResponsibility! !

!DirectoryEntry methodsFor: 'services' stamp: 'cmm 11/30/2008 23:48'!
services
	"Answer the same collection of SimpleServiceEntry's accessed by the FileList."
	^ FileList itemsForFile: self fullName! !

!DirectoryEntry methodsFor: 'private-initialization' stamp: 'cmm 9/13/2007 16:55'!
attributes
	^ attributes ifNil: [ attributes _ Dictionary new ]! !

!DirectoryEntry methodsFor: 'private-initialization' stamp: 'cmm 12/21/2007 16:04'!
setDirectory: aFileOrServerDirectory 
	"Set only my (containing) directory.  This is only needed because I couldn't factor ServerDirectory class>>#parseFTPEntry: to the instance-side (because HTTPClient utility uses it).  Therefore, they pass a nil and then set my 'directory' immediately after.."
	directory := aFileOrServerDirectory! !

!DirectoryEntry methodsFor: 'private-initialization' stamp: 'cmm 9/12/2007 17:20'!
setDirectory: aFileDirectoryOrServerDirectory name: name0  creationTime: creationTime0  modificationTime: modificationTime0 fileSize: fileSize0
	directory := aFileDirectoryOrServerDirectory.
	name := name0.
	creationTime := creationTime0.
	modificationTime := modificationTime0.
	fileSize := fileSize0! !


!DirectoryEntry class methodsFor: 'file list services' stamp: 'cmm 3/25/2009 15:17'!
services
	"Answer the same collection of SimpleServiceEntry's accessed by the FileList."
	self subclassResponsibility! !

!DirectoryEntry class methodsFor: 'instance creation' stamp: 'cmm 9/12/2007 17:21'!
directory: aFileDirectoryOrServerDirectory name: name0 creationTime: creationTime modificationTime: modificationTime fileSize: fileSize 
	^ self new 
		setDirectory: aFileDirectoryOrServerDirectory
		name: name0
		creationTime: creationTime
		modificationTime: modificationTime
		fileSize: fileSize! !

!DirectoryEntry class methodsFor: 'instance creation' stamp: 'cmm 9/12/2007 17:23'!
fromArray: array directory: aFileDirectoryOrServerDirectory 
	| entryType |
	entryType _ (array at: 4) 
		ifTrue: [ DirectoryEntryDirectory ]
		ifFalse: [ DirectoryEntryFile ].
	^ entryType 
		directory: aFileDirectoryOrServerDirectory
		name: (array at: 1)
		creationTime: (array at: 2)
		modificationTime: (array at: 3)
		fileSize: (array at: 5)! !


!DirectoryEntryDirectory methodsFor: 'convert' stamp: 'cmm 9/12/2007 17:40'!
asFileDirectory
	"Answer a FileDirectory representing the same directory I represent."
	^ self containingDirectory in: [ : cd | cd on: (cd fullNameFor: self name) ]! !

!DirectoryEntryDirectory methodsFor: 'file operations' stamp: 'cmm 10/5/2007 17:41'!
copyHere: aDirectoryEntryFile
	"Copy aDirectoryFileEntry, which represents a file, to the directory I represent."
	^ self asFileDirectory copyHere: aDirectoryEntryFile! !

!DirectoryEntryDirectory methodsFor: 'testing' stamp: 'cmm 9/13/2007 12:24'!
isDirectory
	"whether this entry represents a directory, it does."
	^ true! !


!DirectoryEntryDirectory class methodsFor: 'as yet unclassified' stamp: 'cmm 3/25/2009 15:17'!
services
	"Answer the same collection of SimpleServiceEntry's accessed by the FileList."
	^ FileList itemsForDirectory: self fullName! !


!DirectoryEntryFile methodsFor: 'file operations' stamp: 'cmm 9/27/2007 11:49'!
copyTo: targetDirectory 
	"Make a copy of me in targetDirectory.  targetDirectory can be a FileDirectory, ServerDirectory or a DirectoryEntryDirectory.  If a file with my name already exists in targetDirectory, signal a FileExistsException."
	^ targetDirectory copyHere: self! !

!DirectoryEntryFile methodsFor: 'file operations' stamp: 'cmm 9/26/2007 16:46'!
delete
	directory deleteFileNamed: self name! !

!DirectoryEntryFile methodsFor: 'stream access' stamp: 'cmm 10/5/2007 12:13'!
contentsFrom: startPosition to: endPosition 
	"Answer my contents from startPosition to endPosition."
	^ FileStream 
		detectFile: [ self readStream ]
		do: 
			[ : stream | 
			stream
				position: startPosition ;
				next: endPosition - startPosition + 1 ]! !

!DirectoryEntryFile methodsFor: 'stream access' stamp: 'cmm 9/27/2007 16:29'!
contentsTo: endPosition 
	"Answer my contents up to endPosition."
	^ self 
		contentsFrom: 0
		to: endPosition! !

!DirectoryEntryFile methodsFor: 'stream access' stamp: 'cmm 2/19/2009 14:26'!
fileNamed: aString do: oneArgBlock 
	"A convenient accessing method for accessing a file without having to open and close yourself.  The oneArgBlock argument is the fileStream."
	| fileStream |
	fileStream := self fileNamed: aString.
	fileStream ifNil: [ Error signal: 'Could not open ' , aString , ' in use?' ].
	^ [ oneArgBlock value: fileStream ] ensure: [ fileStream close ]! !

!DirectoryEntryFile methodsFor: 'stream access' stamp: 'cmm 10/5/2007 12:15'!
readStream
	"Answer a FileStream on my contents that can be read, but not written."
	^ StandardFileStream readOnlyFileNamed: (directory fullNameFor: self name)! !

!DirectoryEntryFile methodsFor: 'stream access' stamp: 'cmm 9/26/2007 13:15'!
readWriteStream
	"Answer a FileStream on my contents that can be read and written."
	^ directory readOnlyFileNamed: self name! !

!DirectoryEntryFile methodsFor: 'testing' stamp: 'cmm 9/13/2007 12:24'!
isDirectory
	"whether this entry represents a directory, it does not."
	^ false! !


!DirectoryEntryFile class methodsFor: 'file list services' stamp: 'cmm 3/25/2009 15:17'!
services
	"Answer the same collection of SimpleServiceEntry's accessed by the FileList."
	^ FileList itemsForFile: self fullName! !


!FileDirectory methodsFor: 'enumeration' stamp: 'cmm 8/9/2007 16:47'!
directoryNames
	"Return a collection of names for the subdirectories of this directory."
	"FileDirectory default directoryNames"

	^ (self entries select: [:entry | entry isDirectory ])
		collect: [:entry | entry name]
! !

!FileDirectory methodsFor: 'enumeration' stamp: 'cmm 9/19/2007 16:02'!
directoryTreeDo: oneArgBlock 
	"For each file and directory in my tree, value oneArgBlock with an OrderedCollection of the path of DirectoryEntry's leading to the current node.  The first element in the collection will be the DirectoryEntryDirectory for myself, the last is the currentNode (a DirectoryEntry)."
	|myEntry|
	myEntry _ OrderedCollection with: self directoryEntry.
	oneArgBlock value: myEntry.
	self 
		directoryTreeDo: oneArgBlock
		entries: myEntry! !

!FileDirectory methodsFor: 'enumeration' stamp: 'cmm 8/10/2007 12:14'!
entries
	"Return a collection of DirectoryEntry's for the files and directories in this directory.  See primLookupEntryIn:index: for further details."
	"FileDirectory default entries"
	^ self directoryContentsFor: pathName
! !

!FileDirectory methodsFor: 'enumeration' stamp: 'cmm 8/9/2007 16:46'!
fileAndDirectoryNames
	"FileDirectory default fileAndDirectoryNames"
	^ self entries collect: [ : entry | entry name ]! !

!FileDirectory methodsFor: 'enumeration' stamp: 'cmm 8/10/2007 12:00'!
fileNames
	"Return a collection of names for the files (but not directories) in this directory."
	"FileDirectory default fileNames"
	^ (self entries reject: [ : t1 | t1 isDirectory ]) 
		collect: [ : t1 | t1 name ]! !

!FileDirectory methodsFor: 'file operations' stamp: 'cmm 10/5/2007 17:40'!
copyHere: aDirectoryEntryFile 
	"Copy aDirectoryEntryFile, which represents a file, to the directory I represent."
	aDirectoryEntryFile readStream in: 
		[ : readStream | 
		[ self 
			putFile: readStream
			named: aDirectoryEntryFile name ] ensure: [ readStream close ] ]! !

!FileDirectory methodsFor: 'file operations' stamp: 'cmm 4/4/2007 12:49'!
deleteLocalFiles
	"Delete the local files in this directory."
	self fileNames do: [ : fn | self deleteFileNamed: fn ]! !

!FileDirectory methodsFor: 'file status' stamp: 'cmm 4/4/2007 16:33'!
entryAt: fileName ifAbsent: aBlock
	"Find the entry with local name fileName and answer it.
	If not found, answer the result of evaluating aBlock."
	| comparisonBlock |
	self isCaseSensitive
		ifTrue: [comparisonBlock _ [:entry | entry name = fileName]]
		ifFalse: [comparisonBlock _ [:entry | entry name sameAs: fileName]].
	^ self entries detect: comparisonBlock ifNone: [aBlock value]! !

!FileDirectory methodsFor: 'private' stamp: 'cmm 8/9/2007 17:15'!
directoryContentsFor: fullPath
	"Return a collection of directory entries for the files and directories in the directory with the given path. See primLookupEntryIn:index: for further details."
	"FileDirectory default directoryContentsFor: ''"

	| entries index done entryArray f |
	entries := OrderedCollection new: 200.
	index := 1.
	done := false.
	f := fullPath asVmPathName.
	[done] whileFalse: [
		entryArray := self primLookupEntryIn: f index: index.
		#badDirectoryPath = entryArray ifTrue: [
			^(InvalidDirectoryError pathName: pathName asSqueakPathName) signal].
		entryArray == nil
			ifTrue: [done := true]
			ifFalse: [entries addLast: (DirectoryEntry fromArray: entryArray directory: self)].
		index := index + 1].

	^ entries asArray collect: [:s | s convertFromSystemName].
! !

!FileDirectory methodsFor: 'private' stamp: 'cmm 12/21/2007 15:48'!
directoryTreeDo: oneArgBlock entries: entriesCollection 
	"Value oneArgBlock with the path (an OrderedCollection of FileDirectory's) to each DirectoryEntry and the DirectoryEntry itself."
	self entries do: 
		[ : each | 
		entriesCollection add: each.
		oneArgBlock value: entriesCollection.
		each isDirectory ifTrue: 
			[ | subdir |
			subdir _ each asFileDirectory.
			subdir 
				directoryTreeDo: oneArgBlock
				entries: entriesCollection ].
		entriesCollection removeLast ]! !


!AcornFileDirectory methodsFor: 'private' stamp: 'cmm 9/12/2007 17:23'!
directoryContentsFor: fullPath 
	"Return a collection of directory entries for the files and directories in 
	the directory with the given path. See primLookupEntryIn:index: for 
	further details."
	"FileDirectory default directoryContentsFor: ''"

	| entries extraPath |
	entries := super directoryContentsFor: fullPath.
	fullPath isNullPath
		ifTrue: [
			"For Acorn we also make sure that at least the parent of the current dir 
			is added - sometimes this is in a filing system that has not been (or 
			cannot be) polled for disc root names"
			extraPath := self class default containingDirectory.
			"Only add the extra path if we haven't already got the root of the current dir in the list"
			entries detect: [:ent | extraPath fullName beginsWith: ent name] 
				ifNone: [entries := entries
								copyWith: (DirectoryEntryDirectory
										directory: self
										name: extraPath fullName
										creationTime: 0
										modificationTime: 0
										fileSize: 0)]].
	^ entries
! !


!FileList methodsFor: 'volume list and pattern' stamp: 'cmm 8/22/2007 12:08'!
fileNameFormattedFrom: entry sizePad: sizePad
	"entry is a 5-element array of the form:
		(name creationTime modificationTime dirFlag fileSize)"
	| sizeStr nameStr dateStr |
	nameStr _ entry isDirectory
		ifTrue: [entry name , self folderString]
		ifFalse: [entry name].
	dateStr _ ((Date fromSeconds: entry modificationTime )
					printFormat: #(3 2 1 $. 1 1 2)) , ' ' ,
				(String streamContents: [:s |
					(Time fromSeconds: entry modificationTime \\ 86400)
						print24: true on: s]).
	sizeStr _ entry fileSize asStringWithCommas.
	sortMode = #name ifTrue:
		[^ nameStr , '    (' , dateStr , ' ' , sizeStr , ')'].
	sortMode = #date ifTrue:
		[^ '(' , dateStr , ' ' , sizeStr , ') ' , nameStr].
	sortMode = #size ifTrue:
		[^ '(' , ((sizeStr size to: sizePad) collect: [:i | $ ]) , sizeStr , ' ' , dateStr , ') ' , nameStr].
! !

!FileList methodsFor: 'private' stamp: 'cmm 8/22/2007 12:00'!
listForPatterns: anArray
	"Make the list be those file names which match the pattern."

	| sizePad newList |
	newList _ Set new.
	anArray do: [ :pat | newList addAll: (self entriesMatching: pat) ].
	newList _ (SortedCollection sortBlock: self sortBlock) addAll: newList; yourself.
	sizePad _ (newList inject: 0 into: [:mx :entry | mx max: (entry fileSize)])
					asStringWithCommas size - 1.
	newList _ newList collect: [ :e | self fileNameFormattedFrom: e sizePad: sizePad ].

	volList size = 1 ifTrue:
		["Include known servers along with other desktop volumes" 
		^ newList asArray ,
		(ServerDirectory serverNames collect: [:n | '^' , n , self folderString])].
	^ newList asArray! !


!FileList2 methodsFor: 'volume list and pattern' stamp: 'cmm 9/4/2007 12:12'!
listForPatterns: anArray
	"Make the list be those file names which match the patterns."

	| sizePad newList |
	directory ifNil: [^#()].
	(fileSelectionBlock isKindOf: MessageSend) ifTrue: [
		fileSelectionBlock arguments: {directory entries}.
		newList _ fileSelectionBlock value.
		fileSelectionBlock arguments: #().
	] ifFalse: [
		newList _ Set new.
		anArray do: [ :pat |
			newList addAll: (directory entries select: [:entry | fileSelectionBlock value: entry value: pat]) ].
	].
	newList _ newList asSortedCollection: self sortBlock.
	sizePad _ (newList inject: 0 into: [:mx :entry | mx max: entry fileSize])
					asStringWithCommas size - 1.
	newList _ newList collect: [ :e | self fileNameFormattedFrom: e sizePad: sizePad ].
	^ newList asArray! !


!RemoteString methodsFor: 'accessing' stamp: 'ls 9/4/2003 12:37'!
text 
	"Answer the receiver's string asText if remote files are enabled."
	| theFile |
	(sourceFileNumber == nil or: [(SourceFiles at: sourceFileNumber) == nil]) ifTrue: [^ nil].
	theFile := SourceFiles at: sourceFileNumber.
	theFile position: filePositionHi.
	theFile position > theFile size ifTrue: [
		self error: 'RemoteString past end of file' ].
	^ theFile nextChunkText! !

